/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved.
 * -- VMware Confidential
 * **********************************************************/

/*
 * vm_asm_x86_64.h
 *
 *	x86-64 asm macros
 */

#ifndef _VM_ASM_X86_64_H_
#define _VM_ASM_X86_64_H_

#define INCLUDE_ALLOW_VMM
#define INCLUDE_ALLOW_MODULE
#include "includeCheck.h"

#include "x86.h"

/*
 * This file contains inline assembly routines used by x86_64 code.
 * Currently only vmmon and vmnet will use routines defined in this
 * file.
 * 
 * XXX TODO: All code under _MSC_VER have not been implemented.
 */

#ifdef __GNUC__
/* Checked against the Intel manual and GCC --hpreg */
#define _BUILD_SET_DTR(table)  \
static INLINE void             \
_Set_ ## table ## dt(DTR *dtr) \
{                              \
   __asm__(                    \
      "l" #table "dt %0"       \
      :                        \
      : "m" (*dtr)             \
   );                          \
}

#define _BUILD_SET_DTR64(table)    \
static INLINE void                 \
_Set_ ## table ## dt64(DTR64 *dtr) \
{                                  \
   __asm__(                        \
      "l" #table "dt %0"           \
      :                            \
      : "m" (*dtr)                 \
   );                              \
}

_BUILD_SET_DTR(i)
#define SET_IDT(var) do { \
   _Set_idt(&(var));      \
} while (0)
_BUILD_SET_DTR(g)
#define SET_GDT(var) do { \
   _Set_gdt(&(var));      \
} while (0)
_BUILD_SET_DTR64(i)
#define SET_IDT64(var) do { \
   _Set_idt64(&(var));      \
} while (0)
_BUILD_SET_DTR64(g)
#define SET_GDT64(var) do { \
   _Set_gdt64(&(var));      \
} while (0)


/* Checked against the Intel manual and GCC --hpreg

   volatile because the [IG]DT can change without the compiler knowing it
   (when we use l[ig]dt). */
#define _BUILD_GET_DTR(table)  \
static INLINE void             \
_Get_ ## table ## dt(DTR *dtr) \
{                              \
   __asm__ __volatile__(       \
      "s" #table "dt %0"       \
      : "=m" (*dtr)            \
   );                          \
}

#define _BUILD_GET_DTR64(table)    \
static INLINE void                 \
_Get_ ## table ## dt64(DTR64 *dtr) \
{                                  \
   __asm__ __volatile__(           \
      "s" #table "dt %0"           \
      : "=m" (*dtr)                \
   );                              \
}

_BUILD_GET_DTR(i);
#define GET_IDT(var) do { \
   _Get_idt(&(var));      \
} while (0)
_BUILD_GET_DTR(g);
#define GET_GDT(var) do { \
   _Get_gdt(&(var));      \
} while (0)
_BUILD_GET_DTR64(i);
#define GET_IDT64(var) do { \
   _Get_idt64(&(var));      \
} while (0)
_BUILD_GET_DTR64(g);
#define GET_GDT64(var) do { \
   _Get_gdt64(&(var));      \
} while (0)


/* Checked against the Intel manual and GCC --hpreg */
static INLINE void
SET_LDT(Selector s)
{
   __asm__(
      "lldt %0"
      :
      : "rm" (s)
   );
}


/* Checked against the Intel manual and GCC --hpreg

   volatile because the LDT can change without the compiler knowing it
   (when we use lldt). */
static INLINE Selector
_Get_ldt(void)
{
   Selector result;

   __asm__ __volatile__(
      "sldt %0"
      : "=rm" (result)
   );

   return result;
}

#define GET_LDT(var) do { \
   var = _Get_ldt();      \
} while (0)

/* Not checked against the Intel manual and GCC yet --slava */
#define _BUILD_SET_R(func, reg) \
static INLINE void              \
func(uint64 r)                  \
{                               \
   __asm__(                     \
      "movq %0, %%" #reg        \
      :                         \
      : "r" (r)                 \
      : "memory"                \
   );                           \
}

_BUILD_SET_R(SET_CR0, cr0)
_BUILD_SET_R(SET_CR2, cr2)
_BUILD_SET_R(SET_CR3, cr3)
_BUILD_SET_R(SET_CR4, cr4)

_BUILD_SET_R(SET_DR0, dr0)
_BUILD_SET_R(SET_DR1, dr1)
_BUILD_SET_R(SET_DR2, dr2)
_BUILD_SET_R(SET_DR3, dr3)
_BUILD_SET_R(SET_DR6, dr6)
_BUILD_SET_R(SET_DR7, dr7)


/* Not checked against the Intel manual and GCC yet --slava

   volatile because CRs and DRs can change without the compiler knowing it
   (when there is a page fault, when a breakpoint occurs, and moreover it seems
   there is no way to teach gcc that smsw clobbers cr0 for example). */
#define _BUILD_GET_R(reg)   \
static INLINE uint64        \
_Get_ ## reg(void)          \
{                           \
   uint64 result;           \
                            \
   __asm__ __volatile__(    \
      "movq %%" #reg ", %0" \
      : "=r" (result)       \
   );                       \
                            \
   return result;           \
}

_BUILD_GET_R(cr0)
#define GET_CR0(var) do { \
   var = _Get_cr0();      \
} while (0)
_BUILD_GET_R(cr2)
#define GET_CR2(var) do { \
   var = _Get_cr2();      \
} while (0)
_BUILD_GET_R(cr3)
#define GET_CR3(var) do { \
   var = _Get_cr3();      \
} while (0)
_BUILD_GET_R(cr4)
#define GET_CR4(var) do { \
   var = _Get_cr4();      \
} while (0)

_BUILD_GET_R(dr0)
#define GET_DR0(var) do { \
   var = _Get_dr0();      \
} while (0)
_BUILD_GET_R(dr1)
#define GET_DR1(var) do { \
   var = _Get_dr1();      \
} while (0)
_BUILD_GET_R(dr2)
#define GET_DR2(var) do { \
   var = _Get_dr2();      \
} while (0)
_BUILD_GET_R(dr3)
#define GET_DR3(var) do { \
   var = _Get_dr3();      \
} while (0)
_BUILD_GET_R(dr6)
#define GET_DR6(var) do { \
   var = _Get_dr6();      \
} while (0)
_BUILD_GET_R(dr7)
#define GET_DR7(var) do { \
   var = _Get_dr7();      \
} while (0)


/* Checked against the Intel manual and GCC --hpreg */
#define _BUILD_SET_SEG(func, reg) \
static INLINE void                \
func(Selector s)                  \
{                                 \
   __asm__ __volatile__(          \
      "movw %0, %%" #reg          \
      :                           \
      : "rm" (s)                  \
   );                             \
}

_BUILD_SET_SEG(SET_DS, ds)
_BUILD_SET_SEG(SET_ES, es)
_BUILD_SET_SEG(SET_FS, fs)
_BUILD_SET_SEG(SET_GS, gs)
_BUILD_SET_SEG(SET_SS, ss)


/* Checked against the Intel manual and GCC --hpreg

   volatile because the content of CS can change without the compiler knowing
   it (when we use call gates). */
#define _BUILD_GET_SEG(func, reg) \
static INLINE Selector            \
func(void)                        \
{                                 \
   Selector result;               \
                                  \
   __asm__ __volatile__(          \
      "movw %%" #reg ", %0"       \
      : "=rm" (result)            \
   );                             \
                                  \
   return result;                 \
}

_BUILD_GET_SEG(GET_CS, cs)
_BUILD_GET_SEG(GET_DS, ds)
_BUILD_GET_SEG(GET_ES, es)
_BUILD_GET_SEG(GET_FS, fs)
_BUILD_GET_SEG(GET_GS, gs)
_BUILD_GET_SEG(GET_SS, ss)



/* Checked against the Intel manual and GCC --hpreg */
static INLINE void
SET_TR(Selector s)
{
   __asm__(
      "ltr %0"
      :
      : "rm" (s)
   );
}


/* Checked against the Intel manual and GCC --hpreg

   volatile because the content of TR can change without the compiler knowing
   it (when we use task gates). */
static INLINE Selector
_Get_TR(void)
{
   Selector result;

   __asm__ __volatile__(
      "str %0"
      : "=rm" (result)
   );

   return result;
}

#define GET_TR(var) do { \
   var = _Get_TR();      \
} while (0)


/* Checked against the Intel manual and GCC --hpreg */
static INLINE void
_Set_rflags(uint64 f)
{
   __asm__(
      "push %0" "\n\t"
      "popf"
      :
      : "g" (f)
      : "cc"
   );
}

#define RESTORE_FLAGS _Set_rflags


/* Checked against the Intel manual and GCC --hpreg

   volatile because gcc 2.7.2.3 doesn't know when eflags are modified (it
   seems to ignore the "cc" clobber). gcc 2.95.2 is ok: it optimize 2
   successive calls only if there is no instructions in between. */
static INLINE uint64
_Get_rflags(void)
{
   uint64 result;

   __asm__ __volatile__(
      "pushf"  "\n\t"
      "pop %0"
      : "=rm" (result)
   );

   return result;
}

#define SAVE_FLAGS(var) do { \
   var = _Get_rflags();      \
} while (0)

/* Checked against the Intel manual and GCC --hpreg */
static INLINE void
CLTS(void)
{
   __asm__("clts");
}


/* Beginning of the section whose correctness has NOT been checked */
#define FNCLEX()         __asm__("fnclex" ::);

/* TLB_INVALIDATE_xxx checked  yet --slava */
#define TLB_INVALIDATE_PAGE(_addr) do { \
     __asm__ __volatile__("invlpg %0": :"m" (*(char *) (_addr)):"memory"); \
} while (0)

#define TLB_INVALIDATE_PAGE_OFF_FS(_addr) do { \
     __asm__ __volatile__("fs; invlpg %0": :"m" (*(char *) (_addr)):"memory"); \
} while (0)


#if (defined(VMKERNEL) || defined(VMKERNEL_MODULE)) && defined(VMX86_DEBUG)
/* {ENABLE|CLEAR}_INTERRUPTS defined in vmkernel_dist.h */
#include "vmkernel_dist.h"
#else
#define ENABLE_INTERRUPTS() __asm__ __volatile__ ("sti": : :"memory")
#define CLEAR_INTERRUPTS()  __asm__ __volatile__ ("cli": : :"memory")
#endif
#define RAISE_INTERRUPT(_x)  __asm__ __volatile__("int %0" :: "g" (_x))
#define RETURN_FROM_INT()   __asm__ __volatile__("iret" :: ) 

#define NO_INTERRUPTS_BEGIN()	do { \
                                   uint64 _flags; \
                                   SAVE_FLAGS(_flags); \
                                   CLEAR_INTERRUPTS();

#define NO_INTERRUPTS_END()	   RESTORE_FLAGS(_flags); \
                                } while(0)

#elif _MSC_VER  /* !__GNUC__ */

#define SET_IDT(_idt)    { DTR __dt = (_idt); __asm lidt  __dt }
#define SET_GDT(_gdt)    __asm lgdt  _gdt
#define SET_TR(_tr)      __asm ltr  _tr
#define SET_LDT(_tr)     __asm lldt  _tr

#define GET_IDT(_idt)    __asm sidt _idt
#define GET_GDT(_gdt)    __asm sgdt _gdt
#define GET_TR(_tr)      __asm str  _tr
#define GET_LDT(_tr)     __asm sldt  _tr

#define GET_CR0(_reg)    __asm mov eax, cr0 __asm mov _reg, eax
#define SET_CR0(_reg)    __asm mov eax, _reg __asm mov cr0, eax
#define GET_CR2(_reg)    __asm mov eax, cr2 __asm mov _reg, eax
#define SET_CR2(_reg)    __asm mov eax, _reg __asm mov cr2, eax
#define GET_CR3(_reg)    __asm mov eax, cr3 __asm mov _reg, eax
#define SET_CR3(_reg)    __asm mov eax, _reg __asm mov cr3, eax
/*
 * MSC doesn't seem to like CR4 in __asm statements. We emit
 * the opcode for MOV EAX,CR4 = 0xf020e0 and MOV CR4,EAX = 0xf022e0
 */
#define GET_CR4(_reg) { \
 __asm _emit 0x0f __asm _emit 0x20 __asm _emit 0xe0 \
 __asm mov _reg, eax \
}
#define SET_CR4(_reg)    { \
  __asm mov eax, _reg \
  __asm _emit 0x0f __asm _emit 0x22 __asm _emit 0xe0 \
}


#define GET_DR0(_reg) do { __asm mov eax,dr0 __asm mov _reg,eax } while (0)
#define SET_DR0(_reg) do { __asm mov eax,_reg __asm mov dr0,eax } while (0)
#define GET_DR1(_reg) do { __asm mov eax,dr1 __asm mov _reg,eax } while (0)
#define SET_DR1(_reg) do { __asm mov eax,_reg __asm mov dr1,eax } while (0)
#define GET_DR2(_reg) do { __asm mov eax,dr2 __asm mov _reg,eax } while (0)
#define SET_DR2(_reg) do { __asm mov eax,_reg __asm mov dr2,eax } while (0)
#define GET_DR3(_reg) do { __asm mov eax,dr3 __asm mov _reg,eax } while (0)
#define SET_DR3(_reg) do { __asm mov eax,_reg __asm mov dr3,eax } while (0)
#define GET_DR6(_reg) do { __asm mov eax,dr6 __asm mov _reg,eax } while (0)
#define SET_DR6(_reg) do { __asm mov eax,_reg __asm mov dr6,eax } while (0)
#define GET_DR7(_reg) do { __asm mov eax,dr7 __asm mov _reg,eax } while (0)
#define SET_DR7(_reg) do { __asm mov eax,_reg __asm mov dr7,eax } while (0)


#define CLTS()           __asm clts

#define FNCLEX()         __asm fnclex

#define TLB_INVALIDATE_PAGE(_addr) {  \
	 void *_a = (_addr); \
     __asm mov eax, _a __asm invlpg [eax] \
}

#define TLB_INVALIDATE_PAGE_OFF_FS(_addr) { \
	uint32 __a = (uint32) (_addr); \
	__asm mov eax, __a _asm invlpg fs:[eax] \
}


#define ENABLE_INTERRUPTS() { __asm sti }
#define CLEAR_INTERRUPTS()  { __asm cli }

#define RAISE_INTERRUPT(_x)  {__asm int _x }
#define RETURN_FROM_INT()   {__asm iretd }
   

#define SAVE_FLAGS(x) { \
     __asm pushfd __asm pop eax __asm mov x, eax \
}
   
#define RESTORE_FLAGS(x) { \
     __asm push x __asm popfd\
}



static INLINE void SET_DS(Selector val)
{
   __asm mov ax, val
   __asm mov ds, ax
}

static INLINE void SET_ES(Selector val)
{
   __asm mov ax, val 
   __asm mov es, ax
}

static INLINE void SET_FS(Selector val)
{
   __asm mov ax, val 
   __asm mov fs, ax
}

static INLINE void SET_GS(Selector val)
{
   __asm mov ax, val 
   __asm mov gs, ax
}

static INLINE Selector GET_FS(void)
{
	Selector _v;
	__asm mov _v,fs
	return _v;
}

static INLINE Selector GET_GS(void)
{
	Selector _v;
	__asm mov _v,gs
	return _v;
}


static INLINE Selector GET_CS(void)
{
	Selector _v;
	__asm mov _v,cs
	return _v;
}

#pragma warning( disable : 4035)

static INLINE uint32  GET_WORD_FROM_FS(uint32 *_addr) { 
	__asm mov eax, _addr 
    __asm mov eax, fs:[eax] 
}

static INLINE uint16  GET_SHORT_FROM_FS(uint16 *_addr) { 
	__asm mov eax, _addr 
    __asm mov ax, fs:[eax] 
}

static INLINE uint8  GET_BYTE_FROM_FS(uint8 *_addr) { 
	__asm mov eax, _addr 
     __asm mov al, fs:[eax] 
}

#pragma warning (default: 4035)

static INLINE void  SET_WORD_FS(uint32 *_addr, uint32 _val) {
    __asm mov eax, _addr 
    __asm mov ebx, _val
    __asm mov fs:[eax], ebx
}

static INLINE void  SET_SHORT_FS(uint32 *_addr, uint16 _val) {
    __asm mov eax, _addr 
    __asm mov bx, _val
    __asm mov fs:[eax], bx
}

static INLINE void  SET_BYTE_FS(uint32 *_addr, uint8 _val) {
    __asm mov eax, _addr 
    __asm mov bl, _val
    __asm mov fs:[eax], bl
}


#else
#error No compiler defined for get/set
#endif /* !__GNUC__ && !_MSC_VER */


#ifdef __GNUC__
/* Checked against the Intel manual and GCC --hpreg

   volatile because the msr can change without the compiler knowing it
   (when we use wrmsr). */
static INLINE uint64 __GET_MSR(int cx)
{
   uint64 msr;
   uint32 low, high;

   __asm__ __volatile__(
      "rdmsr"
      : "=a" (low), "=d" (high)
      : "c" (cx)
   );

   msr = ((uint64)low) | (((uint64)high) << 32);
   return msr;
}
#elif _MSC_VER
#pragma warning( disable : 4035)
static INLINE uint64 __GET_MSR(int input)
{
   __asm push ecx
   __asm mov  ecx, input
   __asm _emit 0x0f __asm _emit 0x32
   __asm pop ecx
}   
#pragma warning (default: 4035)
#else
#error
#endif

#ifdef __GNUC__
static INLINE void __SET_MSR(int cx, uint64 val)
{
   uint32 low, high;
   low = val;
   high = val >> 32;
   __asm__ __volatile__(
      "wrmsr"
      : /* no outputs */			
      : "a" (low), "d"(high),
        "c" (cx)
    );
}
#endif

static INLINE void SET_FS64(uint64 fs64)
{
   __SET_MSR(0xc0000100, fs64);
}

static INLINE void SET_GS64(uint64 gs64)
{
   __SET_MSR(0xc0000101, gs64);
}

static INLINE uint64 GET_FS64(void)
{
   return __GET_MSR(0xc0000100);
}

static INLINE uint64 GET_GS64(void)
{
   return __GET_MSR(0xc0000101);
}


#define START_TRACING() { \
   uint32 flags;          \
   SAVE_FLAGS(flags);     \
   flags |= EFLAGS_TF;    \
   RESTORE_FLAGS(flags);  \
}

#define STOP_TRACING() { \
   uint32 flags;          \
   SAVE_FLAGS(flags);     \
   flags &= ~EFLAGS_TF;    \
   RESTORE_FLAGS(flags);  \
}  

static INLINE Bool
INTERRUPTS_ENABLED(void)
{
   uint32 flags;
   SAVE_FLAGS(flags);
   return ((flags & EFLAGS_IF) != 0);
}

/*
 *-----------------------------------------------------------------------------
 *
 * ChangeAddrSpaceWithoutDelay --
 *   Flushes the TLB and changes the address space by assigning CR3
 *   to the specified value.  Does not include delay for working around
 *   PIII CR3 setting problem (see BUG 6176). Work around is only necessary
 *   if monitor uses global pages. 
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   Flushes TLB and changes the address space.
 *
 *-----------------------------------------------------------------------------
 */
static INLINE void
ChangeAddrSpaceWithoutDelay(uint32 cr3value)
{
   SET_CR3(cr3value);
}

#endif
